# Copyright 2006, TIG
#
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided the above
# copyright notice appear in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#-----------------------------------------------------------------------------
# Name        : VolumeCalculator
#
# Description : A tool to calculate the volume of a selected set of 3D faces
#
# Menu Item   : See menu-section at end on how to include it here: Plugins -> Volume
#
# Context Menu: Volume
#
# Author      : TIG
#
# Usage       : First select a Group or Component that has faces that form a 
#		volume, then use the right-click Context-Menu and choose 'Volume'.
#		If there is the Plugins Menu Item 'Volume' you can aso pick that.
#		(see the end menu section on how to activate that option...)
#
#		A dialog asks for the volume units and the accuracy required.
#		The units can be cu.m, cc, cu.yds, cu.ft, cu.ins, litres, cl, ml, 
#		gallons(UK), gallons(USA), quarts(USA), pints(UK) and pints(USA).  
#		Volumes are given to varying d.p's depending on the units chosen.   
#		The accuracy can be 0.5, 1, 2, 5 or 10% - the more accurate you 
#		make it the longer it takes to process and the more complex the new 
#		volume-group will become.  Small simple forms will report accurately 
#		even with the 10% sampling...
#
#		A group is made with faces and a text-tag. The text-tag is 
#		added 2" above the the apex of the selection's bounding-box.  
#		It shows the volume in the required units using the current 
#		text/font settings.  A possible new layer can be used.  It is 
#		named VOLS-nnnnnnnn (where nnnnnnnn is based on the date/time). 
#		Alternatively you can choose to use any existing layer OR make 
#		your own 'on the fly'.
#
#		It then calculates a series of thin horizontal slices through the 
#		selection's faces to use to calculate the volume enclosed. 
#		There are now also new Layer, Hide/Show-Edges and Colour Options 
#		in v1.7, see below...***
#
#		Before ending it checks if 'Xray Mode' was already switched 'on' at 
#		the start and if so there is no action and it exits.  Otherwise the 
#		original selection is 'Hidden' so you can clearly see the volume-group.  
#		You are then asked in a dialog if you want to leave it 'Hidden'.  
#		If 'Yes' then it exits leaving it 'Hidden'.  If 'No' the original 
#		selection is 'Unhidden' and 'Xray Mode' is then switched 'on' so you 
#		can see the volume-group inside of the original selection and a dialog 
#		asks if you want to leave 'Xray Mode' switched 'on'.  If you answer 
#		'No' the normal view is restored and it exits.  If you answer 'Yes' 
#		it exits leaving 'Xray Mode' switched 'on'.
#
#		If you don't want the slice to be visible you can edit the 
#		volume's group and either delete all of its geometry, leaving 
#		the text-tag in place, or just select all geometry and hide it.
#
#		'Volume' can't be expected to be 100% foolproof !  It treats an open 
#		topped object as if it were filled with water, taking volumes from the 
#		outer 'skin' inwards to work out which are the solid and which are the 
#		hollow volumes.  With a little ingenuity you will be able to contrive
#		a form with complex linked holes punched in, or a vertical array of 
#		thin volumes spaced apart so as to miss the sampling slices - which 
#		because it uses a series of horizontal slices, can produce unexpected 
#		results. Keep volumes simple and build them up in bits for safety...
#
#		***New Options(v1.7) let you choose any layer OR make a new one,
#		hide/show volume's edges and give faces any standard colour -
#		pick popout' type first letter of colour to jump down list -
#		e.g. W for White takes you to Wheat...: 
#		use up/down arrow keys to move through list...
#		or pick any model material listed...
#		
# Date        : 12/2005
#
# Type        : Tool
#
# Verison     :	1.0	17/12/05	First issue.
#		1.1	18/12/05	Minor adjustments to disc definition and text.
#		1.2	18/12/05	Circle def moved into Class (Mac friendly).
#		1.3	19/12/05	Cutting disc made larger to include all bbox.
#		1.4	20/12/05	V5 only support included.
#		1.5	21/12/05	Undo text corrected.
#		1.6	07/01/06	Volume calculation error in non-minZ=0 group fixed.
#		1.7 14/07/07	Layer, Hide/Show-Edges and Colour Options added.
#		1.8 12/11/07	Bugsplat causes investigated and fixed by 2 step undo???
#
#-----------------------------------------------------------------------------
require 'sketchup.rb'
###
class Volume
###---------------------------------------
def Volume::calculate
### version error
if (Sketchup.version.split(".")[0].to_i<5)
     UI.beep
     UI.messagebox("Sorry.  This is Version 5+ program only !")
     return nil
end#if
###
model = Sketchup.active_model
entities = model.active_entities
model.start_operation "Volume"
ss = model.selection
view = model.active_view
  if ss.empty?
     UI.beep
     UI.messagebox("NO Selection !")
     return nil
  end#if
  if ss[1]
     UI.beep
     UI.messagebox("Selection MUST be ONE Group or Component !")
     return nil
  end#if
  if ss[0].typename != "Group" and ss[0].typename != "ComponentInstance"
     UI.beep
     UI.messagebox("Selection is NOT a Group or Component !")
  end#if
###
selected = ss[0]
###
# -------------- set up @units and @percent in def dialog for later...
# -------------- make cutting disc face at xy bounds
###
###
bb = selected.bounds
xmin = bb.min.x
ymin = bb.min.y
xmax = bb.max.x
ymax = bb.max.y
zmin = bb.min.z
zmax = bb.max.z
###
if xmin == xmax or ymin == ymax or zmin == zmax
   UI.beep
   UI.messagebox("The Selection has no Volume !")
   return nil
end#if
###
# ----------- dialog ----------------
### show VCB and status info
  Sketchup::set_status_text(("VOLUME PARAMETERS..." ), SB_PROMPT)
  Sketchup::set_status_text(" ", SB_VCB_LABEL)
  Sketchup::set_status_text(" ", SB_VCB_VALUE)
  return nil if not Volume.dialog ### do dialog...
###
#### ------- slice & z inc ------------------------------
slice = ((zmax-zmin)*@percent)### THE basic slice
increment_z = (zmax-zmin)/(((zmax-zmin)/slice).ceil)
###
apex = Geom.linear_combination(0.5,[xmin,ymin,zmax+2],0.5,[xmax,ymax,zmax+2])
### +2 puts it above top ( for text loc'n )
pnt1 = [xmin,ymin,zmin] ### v1.1
pnt2 = [xmax,ymax,zmin] ### v1.1
ctr = Geom.linear_combination(0.5,pnt1,0.5,pnt2)
rad = (pnt1.distance pnt2)### make a LOT bigger than bb ###v1.3
###
disc = entities.add_group
discentities = disc.entities
disccircle = Volume.circle(ctr,[0,0,-1],rad,24)
discentities.add_face disccircle
###
#--------------------------- reset Z max and min's
###zmax = zmax-(increment_z/2)### v1.6 NOT required ###
zmin = zmin+(increment_z/2)
###
# ------------------------ do each of slice
vol=entities.add_group
volentities=vol.entities
vol.name="Volume-"+@sfix
###
### intersect disc and selected
###
### The intersect_with method is used to intersect entities, a component instance, or group
### with an entities object.
###
### entities.intersect_with recurse, transformation1, entities1, transformation2, hidden, entities2
###
### recurse - true if you want this entities object to be recursed 
###   (intersection lines will be put inside of groups and components within this entities object)
### transformation1 - the transformation for this entities object
### entities1 - the entities (group etc) where you want the intersection lines to appear
### transformation2 - the transformation for entities1
### hidden - true if you want hidden geometry in this entities object to be used in the intersection
###   (it's false in section cuts)
### entities2 - an entities object (or an array of entities ?)
###
### loop through all possible slices - bottom up
n = ((zmax-zmin)/slice).ceil
c = 0
z = increment_z/2 ### move to start v1.6 fix ###
while z < zmax-zmin+increment_z ### v1.6 fix ###
### show VCB and status info
  Sketchup::set_status_text(("MAKING VOLUME SLICE #{c} of #{n.to_i}" ), SB_PROMPT)
  Sketchup::set_status_text(" ", SB_VCB_LABEL)
  Sketchup::set_status_text(" ", SB_VCB_VALUE)
### do cut
  disc.move! Geom::Transformation.new [0,0,z] ### first placing near bottom v1.6
  discentities.intersect_with true, disc.transformation, vol, vol.transformation, true, selected
  z = z+increment_z
  c = c+1
end#while
### delete disc ###v1.8
disc.entities.to_a.each{|e|e.hidden=true if e.valid?}### so disc is hidden
model.commit_operation### so no bugsplat
model.start_operation("Volume...")
disc.erase!###v1.8
### face the slices
for e in volentities
### show VCB and status info
  Sketchup::set_status_text(("MAKING VOLUME SLICES..." ), SB_PROMPT)
  Sketchup::set_status_text(" ", SB_VCB_LABEL)
  Sketchup::set_status_text(" ", SB_VCB_VALUE)
  e.find_faces if e.typename == "Edge"
end#for e
###
faces = []
for f in volentities
   if f.typename == "Face"
      faces = faces.push(f)
   end
end#for f
faces=faces.uniq
###
for face in faces
   face.reverse! if face.normal==[0,0,-1] ### now all face up
end#for face
### now work out which is which...
keptfaces = []
(faces.length-1).times do |this|
   for face in faces
    if face.valid?
      for edgeuse in face.outer_loop.edgeuses
         if not edgeuse.partners[0] ### outermost face
            keptfaces = keptfaces.push(face)
            faces = faces - [face]
            loops = face.loops
            for loop in loops
               for fac in faces
                  if fac.valid? and (fac.outer_loop.edges - loop.edges) == []
                     faces = faces - [fac]
                     fac.erase!### fac abutts kept face so it must be erased...
                  end #if fac
               end #for fac
            end #for loop
         end #if outermost
      end #for edgeuse
    end #if valid
   end #for face
end#times
###
layerVolume=@layer
if layerVolume
  layerVolume=model.layers.add(layerVolume)
  layerVolume.page_behavior=(LAYER_IS_HIDDEN_ON_NEW_PAGES | LAYER_HIDDEN_BY_DEFAULT) 
  layerVolume.visible=true
end#if
vol.layer=layerVolume
### -------------- now find area of all edges etc
colour=@colour
colour=nil if colour=="<Default>"
area = 0
for f in volentities
### show VCB and status info
  Sketchup::set_status_text(("CALCULATING VOLUME..." ), SB_PROMPT)
  Sketchup::set_status_text(" ", SB_VCB_LABEL)
  Sketchup::set_status_text(" ", SB_VCB_VALUE)
  f.material=colour if f.typename=="Face"
  area=(area+f.area) if f.typename=="Face"
  ###f.erase! if f.typename == "Edge" and not f.faces[0]
  f.hidden= true if (f.valid? && f.typename=="Edge") && @hidden=="Yes"
end#for f
###
if area==0
   UI.beep
   UI.messagebox("There is NO Volume !")
   vol.erase! if vol.valid?
   return nil
end#if
# --------------- get volume ----
###
volume=(area*increment_z)### in cubic inches
###
### convert it to required units...
###["cu.m", "cc", "cu.yds", "cu.ft", "cu.ins", "litres", "cl", "ml", "gallons(UK)", "gallons(USA)", "quarts(USA)", "pints(UK)", "pints(USA)"]
volumetxt = volume.to_s
volumetxt = ((volume* 0.000016387064 *1000).round.to_f/1000).to_s if @units == "cu.m"
volumetxt = ((volume* 16.387064).round.to_i).to_s if @units == "cc"
volumetxt = ((volume* 0.000021433471 *1000).round.to_f/1000).to_s if @units == "cu.yds"
volumetxt = ((volume* 0.000578703704 *100).round.to_f/100).to_s if @units == "cu.ft"
volumetxt = ((volume *10).round.to_f/10).to_s if @units == "cu.ins"
volumetxt = ((volume* 0.016387064 *1000).round.to_f/1000).to_s if @units == "litres"
volumetxt = ((volume* 1.6387064 *10).round.to_f/10).to_s if @units == "cl"
volumetxt = ((volume* 16.387064).round.to_i).to_s if @units == "ml"
volumetxt = ((volume* 0.003604750773 *100).round.to_f/100).to_s if @units == "gallons(UK)"
volumetxt = ((volume* 0.004329125301 *100).round.to_f/100).to_s if @units == "gallons(USA)"
volumetxt = ((volume* 0.017316501203 *10).round.to_f/10).to_s if @units == "quarts(USA)"
volumetxt = ((volume* 0.028838006186 *10).round.to_f/10).to_s if @units == "pints(UK)"
volumetxt = ((volume* 0.034633002405 *10).round.to_f/10).to_s if @units == "pints(USA)"
# add text
t=volentities.add_text((volumetxt+" "+@units),apex)
t.layer=layerVolume
###
### show VCB and status info
  Sketchup::set_status_text((" " ), SB_PROMPT)
  Sketchup::set_status_text(" ", SB_VCB_LABEL)
  Sketchup::set_status_text(" ", SB_VCB_VALUE)
### unselected...
model.selection.clear
###
### hide selected
xray = 0
if not model.rendering_options["ModelTransparency"] ### NOT xray mode
   selected.hidden= true
   if UI.messagebox("Do you want to leave the Original Selection 'Hidden' for now ?  ",MB_YESNO,"Volume Hidden ?") == 7 ### 6=YES 7=NO
      xray = 1
      selected.hidden= false
   end#if
end#if
### xray ?
xrayq= 0
if not model.rendering_options["ModelTransparency"]
   xrayq = 1
   model.rendering_options["ModelTransparency"]=true if xray == 1
end#if
###
### ---- undo xray IF was not on already
if xray == 1 and xrayq == 1
  model.rendering_options["ModelTransparency"]=false if UI.messagebox("Do you want to leave 'Xray Mode' on for now ?  ",MB_YESNO,"Volume Xray ?")==7 ###7=NO
end#if
###
### update view
view.invalidate
# ---------------------- Close/commit group
model.commit_operation
#-----------------------
end#def
###################################################---------------------
def Volume::dialog
### get units and accuracy etc
   units = ["cu.m", "cc", "cu.yds", "cu.ft", "cu.ins", "litres", "cl", "ml", "gallons(UK)", "gallons(USA)", "quarts(USA)", "pints(UK)", "pints(USA)"].join('|')
   percents = ["0.5 %", "1 %", "2 %", "5 %", "10 %"].join('|')
   mlayers=Sketchup.active_model.layers
   layers=[]
   mlayers.each{|e|layers.push e.name}
   dlayer=layers[0]
   layers=layers-[dlayer]
   layers.sort!
   #----------- sort possible special Layer
   @sfix=Time::now.to_i.to_s[3..-1]
   layerVolume=("VOLS-"+@sfix)
   @makelayer="<Make New Layer>"
   layers=[layerVolume]+[@makelayer]+[dlayer]+layers
   layers.uniq!
   layers=layers.join('|')
   hidden="Yes|No"
   mcolours=Sketchup.active_model.materials
   colours=[]
   mcolours.each{|e|colours.push e.name}
   colours.sort!
   colours=colours+["<Default>"]+(Sketchup::Color.names.sort!)
   colours.uniq!
   colours=colours.join('|')
   prompts = ["Units: ","Accuracy: ","Layer: ","Hide Edges ? : ","Colour: "]
   title = "Volume Parameters"
   @units = "cu.m" if not @units
   @dpercent = "5 %" if not @dpercent
   @layer=layerVolume
   @hidden="Yes" if not @hidden
   @colour="<Default>" if not @colour
   values = [@units,@dpercent,@layer,@hidden,@colour]
   popups = [units,percents,layers,hidden,colours]
   results = inputbox(prompts,values,popups,title)
   return nil if not results
### do processing of results
@units,@dpercent,@layer,@hidden,@colour=results
@percent = 0.005 if @dpercent == "0.5 %"
@percent = 0.01 if @dpercent == "1 %"
@percent = 0.02 if @dpercent == "2 %"
@percent = 0.05 if @dpercent == "5 %"
@percent = 0.1 if @dpercent == "10 %"
### make layer dialog
if @layer==@makelayer
   results2=inputbox(["New Volume's Layer Name: "],["VOLS-????"],"New Layer Name")
   if results2
      @layer=results2[0]
   else
      @layer=nil
   end#if
end#if
###
true
###
end #def dialog
##################
# --- Function for generating points on a circle
def Volume::circle(center,normal,radius,numseg)
    # Get the x and y axes
    axes = Geom::Vector3d.new(normal).axes
    center = Geom::Point3d.new(center)
    xaxis = axes[0]
    yaxis = axes[1]
    xaxis.length = radius
    yaxis.length = radius
    # compute the points
    da = (Math::PI * 2) / numseg
    pts = []
    for i in 0...numseg do
        angle = i * da
        cosa = Math.cos(angle)
        sina = Math.sin(angle)
        vec = Geom::Vector3d.linear_combination(cosa,xaxis,sina,yaxis)
        pts.push(center + vec)
    end
    # close the circle
    pts.push(pts[0].clone)
    pts
end#def
###########################
#-----------------------
end#class
#-----------------------
#--------- menu -----------------------------
if( not file_loaded?("VolumeCalculator.rb") )
   ###add_separator_to_menu("Plugins")
   ###UI.menu("Plugins").add_item("Volume") { Volume.calculate }
### remove ### in front of above 2 lines if to display in Plugins...
   UI.add_context_menu_handler do | menu |
      if (Sketchup.active_model.selection[0].typename == "Group" or Sketchup.active_model.selection[0].typename == "ComponentInstance")
         menu.add_separator
         menu.add_item("Volume") { Volume.calculate }
      end #if ok
   end #do menu
end#if
file_loaded("VolumeCalculator.rb")
#---------------------------------
